# Oliver Gordon, 2019

from nOmicron.mate import objects as mo

# from selenium import webdriver
# import chromedriver_autoinstaller
#
#
# chromedriver_autoinstaller.install()  # Check if the current version of chromedriver exists
#                                       # and if it doesn't exist, download it automatically,
#                                       # then add chromedriver to path
#
# driver = webdriver.Chrome()
# driver.get("http://www.python.org")

def set_gap_voltage(voltage):
    mo.gap_voltage_control.Voltage(voltage)


def approach(v_gap, cautiousness):
    modes = {"Super Cautious": {"Loop_Gain_I": 3, "I_Setpoint": 50e-12},
             "Cautious": {"Loop_Gain_I": 3, "I_Setpoint": 50e-12},
             "Normal": {"Loop_Gain_I": 3, "I_Setpoint": 50e-12},
             "YOLO": {"Loop_Gain_I": 3, "I_Setpoint": 50e-12}}

    mo.regulator.Preamp_Range_1(0)
    mo.regulator.Loop_Gain_1_I(modes[cautiousness["Loop_Gain_I"]])
    mo.regulator.Setpoint_1(modes[cautiousness["I_Setpoint"]])
    mo.gap_voltage_control.Voltage(v_gap)

    print("Ready to beginning approach...")
    while not mo.regulator.Setpoint_Detected():
        pass
    print("Approach finished.")



#   Code up to mate4dummies 0.4.2 (at time of branch) is Copyright © Stephan Zevenhuizen
#   MATE Objects, (10-05-2018).
#   MATE for Dummies, (20-08-2018).

"""
MATE for Dummies. MATE Scripting System in Python
"""

__version__ = '1.0'

__all__ = ['MATE','objects']

from .mate import *



# Oliver Gordon, 2019

import ctypes
from nOmicron.mate import objects as mo

from nOmicron.microscope import IO


def view_channel_properties(channel_name):
    # TODO: Rewrite!
    def read_prop(obj, prop):
        out = getattr(obj, prop)()
        print(prop + ':', out)

    channel_props = ['Enable']
    clock_props = ['Enable', 'Period', 'Samples']
    experiment_props = ['Bricklet_Written', 'Name', 'Result_File_Name',
                        'Result_File_Path', 'State']
    gap_voltage_control_props = ['Preamp_Range', 'Voltage']
    regulator_props = ['Enable_Z_Offset_Slew_Rate', 'Feedback_Loop_Enabled',
                       'Loop_Gain_1_I', 'Loop_Gain_2_I', 'Preamp_Range_1',
                       'Preamp_Range_2', 'Setpoint_1', 'Setpoint_2', 'Z_Offset',
                       'Z_Offset_Slew_Rate', 'Z_Out']
    view_props = ['Cycle_Count', 'Data_Size', 'Deliver_Data', 'Run_Count']
    xy_scanner_props = ['Angle', 'Area', 'Enable_Drift_Compensation', 'Lines',
                        'Offset', 'Plane_X_Slope', 'Plane_Y_Slope', 'Points',
                        'Raster_Time', 'Return_To_Stored_Position',
                        'Store_Current_Position', 'Target_Position',
                        'Trigger_Execute_At_Target_Position', 'XY_Position_Report',
                        'X_Drift', 'X_Retrace', 'X_Retrace_Trigger',
                        'X_Trace_Trigger', 'Y_Drift', 'Y_Retrace',
                        'Y_Retrace_Trigger', 'Y_Trace_Trigger']

    obj_names = ['channel', 'clock', 'experiment', 'gap_voltage_control',
                 'regulator', 'view', 'xy_scanner', 'spectroscopy']
    IO.connect()
    IO.enable_channel(channel_name)
    for obj_name in obj_names:
        obj = getattr(mo, obj_name)
        print()
        print('Object ' + obj_name + ':')
        print('--------------------------------------------------------------')
        for prop in eval(obj_name + '_props'):
            read_prop(obj, prop)

    mo.mate.disconnect()


def check_response_code(string_to_check, object_type):
    """Read the response code described by the Mate manual. e.g. View.I_V.Deliver_Data"""

    class String(ctypes.Structure):
        _fields_ = [('length', ctypes.c_int),
                    ('text', ctypes.c_char * 256)]

    bt = bytearray()
    bt.extend(string_to_check.encode())
    string_to_check = bytes(bt)

    lib_mate = mo.mate.lib_mate
    b = ctypes.c_char()
    i = ctypes.c_int()
    e = ctypes.c_int()
    d = ctypes.c_double()
    s = String(255)
    p_s = ctypes.pointer(s)
    d1 = ctypes.c_double()
    d2 = ctypes.c_double()

    response = dict()
    response_strings = {v: k for k, v in mo.mate.rcs.items()}

    if object_type == "bool":
        response["response_code"] = lib_mate.getBooleanProperty(string_to_check, -1, ctypes.byref(b))
        response["response_value"] = bool(ord(b.value))
    elif object_type == "int":
        response["response_code"] = lib_mate.getIntegerProperty(string_to_check, -1, ctypes.byref(b))
        response["response_value"] = i.value
    elif object_type == "str":
        response["response_code"] = lib_mate.getStringPropertyByDesc(string_to_check, -1, ctypes.byref(p_s))
        response["response_value"] = s.text[:].decode()
    elif object_type == "enum":
        response["response_code"] = lib_mate.getEnumProperty(string_to_check, -1, ctypes.byref(e))
        response["response_value"] = e.value
    elif object_type == "double":
        response["response_code"] = lib_mate.getDoubleProperty(string_to_check, -1, ctypes.byref(d))
        response["response_value"] = d.value
    elif object_type == "pair":
        response["response_code"] = lib_mate.getPairProperty(string_to_check, -1, ctypes.byref(d1), ctypes.byref(d2))
        response["response_value"] = (d1.value, d2.value)
    else:
        raise KeyError("Not a gettable type")

    response["response_string"] = response_strings[response["response_code"]]

    return response



#   Copyright © Oliver Gordon, 2019

"""
nOmicron package/API for automatically control of Matrix through Python
"""

__version__ = '1.0'

__all__ = ['mate', 'microscope', 'utils']
name = "nOmicron"

from .mate import *
from .microscope import *
from .utils import *



class MatrixUnexpectedResponseError(Exception):
    """RMT_UNEXPECTEDRESPONSE"""

    def __init__(self, message="An unexpected response was returned."):
        self.message = message
        super().__init__(self.message)


class MateLibNotLoadableError(Exception):
    """RMT_LIBNOTLOADABLE"""

    def __init__(self, message=f"Cannot load the MATE remote access API"):
        self.message = message
        super().__init__(self.message)


class MatrixCriticalHardwareFailureError(Exception):
    """RMT_INTERNALFAILURE"""

    def __init__(self, message="The Matrix CU has reported a hardware failure. \n "
                               "Restarting the CU and Matrix should fix this.."):
        self.message = message
        super().__init__(self.message)


class MatrixNotInitialisedError(Exception):
    """RMT_NOSERVER"""

    def __init__(self, message=f"Matrix must be open and initialised before calling IO.connect()?"):
        self.message = message
        super().__init__(self.message)


class MatrixIncompatibleProtocolError(Exception):
    """RMT_INCOMPATIBLEPROTOCOL"""

    def __init__(self, message=f"There is an incompatability between Matrix.exe and the remote access API. \n "
                               f"This may be due to a 32/64 bit iocompatiblity. \n"
                               f"Alternatively, if multiple versions of Matrix are installed, inconsistent naming \n"
                               f"between versions during installation can cause \n"
                               f"nOmicron to point to incorrect Matrix paths."):
        self.message = message
        super().__init__(self.message)


class MatrixUnsupportedReqError(Exception):
    """RMT_UNSUPPORTEDREQ"""

    def __init__(self, message=f"Req is not supported"):
        self.message = message
        super().__init__(self.message)


class MatrixUnsupportedOperationError(Exception):
    """RMT_UNKNOWNOBJECT, RMT_UNKNOWNPROPERTY"""

    def __init__(self, message=f"Requested operation is not supported \n"
                               f"in the current experiment/hardware configuration"):
        self.message = message
        super().__init__(self.message)


class MatrixInvalidDataTypeError(Exception):
    """RMT_INVALIDTYPE"""

    def __init__(self, message=f"The requested object type is not compatible with the requested operation"):
        self.message = message
        super().__init__(self.message)


class MatrixRejectedError(Exception):
    """RMT_REJECTED"""

    def __init__(self, message=f"The requested operation has been rejected"):
        self.message = message
        super().__init__(self.message)


class MatrixParameterOutOfRangeWarning(Warning):
    def __init__(self, message="Requested parameter value is outside of tolerable range. \n "
                               "Matrix may behave oddly or die "):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return repr(self.message)



from selenium import webdriver
import chromedriver_autoinstaller
from selenium.webdriver.chrome.options import Options
import subprocess

from nOmicron.utils.errors import MatrixUnsupportedOperationError

if "Received = 1" not in subprocess.check_output("ping 10.0.42.2 -n 1", shell=True).decode("utf-8"):
    raise MatrixUnsupportedOperationError("Remote black box control is unavailable in this hardware configuration. \n"
                                          "Check that http://10.0.42.2 can be loaded")
chromedriver_autoinstaller.install()

options = Options()
options.add_argument("--headless")
driver = webdriver.Chrome(options=options)
driver.get("http://10.0.42.2/")


def x_plus():
    driver.find_element_by_name("F1p").click()


def x_minus():
    driver.find_element_by_name("F1m").click()


def y_plus():
    driver.find_element_by_name("F2p").click()


def y_minus():
    driver.find_element_by_name("F2m").click()


def z_plus():
    driver.find_element_by_name("F3p").click()


def z_minus():
    driver.find_element_by_name("F3m").click()


def up():
    driver.find_element_by_name("UP").click()


def down():
    driver.find_element_by_name("DOWN").click()


def fx():
    driver.find_element_by_name("FX").click()


def backward():
    driver.find_element_by_name("BWD").click()


def forward():
    driver.find_element_by_name("FWD").click()


def auto_approach():
    driver.find_element_by_name("AUTO").click()


def auto_approach():
    driver.find_element_by_name("AUTO").click()



# Oliver Gordon, 2019

import warnings

import nOmicron.mate.objects as mo
import numpy as np
from nOmicron.microscope import IO
from nOmicron.utils.plotting import plot_xy
from tqdm import tqdm
from time import sleep

def set_gap_voltage(voltage):
    mo.gap_voltage_control.Voltage(voltage)


def set_scan_position(xy_pos=None, width_height=None, angle=None):
    mo.xy_scanner.Enable_Scan(False)

    if width_height[0] != width_height[1]:
        mo.xy_scanner.Width_Height_Constrained(False)

    if width_height is not None:
        mo.xy_scanner.Width(width_height[0])
        mo.xy_scanner.Height(width_height[1])

    if xy_pos is not None:
        mo.xy_scanner.X_Offset(xy_pos[0])
        mo.xy_scanner.Y_Offset(xy_pos[1])

    if angle is not None:
        mo.xy_scanner.Angle(angle)

    mo.xy_scanner.Enable_Scan(True)


def set_points_lines(points=None, lines=None):
    """
    Sets the number of points and lines of the scan.

    Parameters
    ----------
    points : int or None
        Number of points
    lines : int or None
        Number of lines

    Examples
    --------
    Set both points and lines to 150
    >>> set_points_lines(150)

    Set both points and lines to 150
    >>> set_points_lines(points=150, lines=150)

    Set points and lines to different amounts
    >>> set_points_lines(110, 130)
    """
    mo.experiment.stop()
    mo.xy_scanner.Points_Lines_Constrained(True)
    if points != lines:
        if points is not None and lines is not None:
            mo.xy_scanner.Points_Lines_Constrained(False)
        else:
            lines = points

    mo.xy_scanner.Points(points)
    mo.xy_scanner.Lines(lines)


def set_scan_speed(scan=None, move=None, speed_adjust_mode="Frequency"):
    """
    Sets the scan and move t-raster elements

    Parameters
    ----------
    scan : float or None
        Time span between two adjacent raster points, in seconds
    move : float or None
        Time span between movements of the y scan, in seconds
    speed_adjust_mode : str
        The adaption method to keep constant when changing the adjust the scan raster/speed time.
        Must be either ["Frequency", "Speed"]. Default is "Frequency".

    Examples
    --------
    Set both scan and move to 150ms
    >>> set_scan_speed(150e-3)

    Set both points and lines to 150ms
    >>> set_scan_speed(points=150e-3, lines=150e-3)

    Set points and lines to different amounts
    >>> set_scan_speed(110e-3, 130e-3)
    """
    speed_modes = {"Frequency": 1, "Speed": 2}
    if speed_adjust_mode not in speed_modes.values():
        raise ValueError(f"Must be one of {speed_modes.values()}")
    mo.xy_scanner.Speed_Adaption(speed_modes[speed_adjust_mode])

    mo.xy_scanner.Move_Raster_Time_Constrained(True)
    if scan != move:
        if scan is not None and move is not None:
            mo.xy_scanner.Move_Raster_Time_Constrained(False)
        else:
            move = scan

    mo.xy_scanner.Raster_Time(scan)
    mo.xy_scanner.Move_Raster_Time(move)


def get_xy_scan(channel_name, x_direction, y_direction, num_lines='all', mode='new', return_filename=False):
    """
    Perform and get an xy scan

    Parameters
    ----------
    channel_name : str
        The channel to acquire from, e.g. Z, I, Aux1, Aux2
    x_direction : str
        Must be one of 'Forward', 'Backward', or 'Forward-Backward'
    y_direction : str
        Must be one of 'Up' or 'Up-Down'
    num_lines : int or str
        Number of lines to get. If an int, must be less than the number of lines in the scanner window.
        Default is 'all'
    mode : str, optional
        Must be one of ['new', 'pause', 'continue']. Default is 'new'
    return_filename : bool, optional
        If the full file name of the scan should be returned along with the data. Default is False

    Returns
    -------
    xydata : ndarray
        Max 4 dimensions (y_up/y_down, x_up/x_down, x, y). First two dimensions will only appear if
        y_direction is "Up-Down" and x_direction is "Forward-Backward" - will be missing as appropriate

    Examples
    --------
    >>> from nOmicron.microscope import IO
    >>> from nOmicron.utils.plotting import plot_xy
    >>> IO.connect()
    >>> xydata = get_xy_scan("Z", x_direction="Forward", y_direction="Up-Down")
    >>> plot_xy(xydata, pixel_scale=mo.xy_scanner.Width() * 1e9 / mo.xy_scanner.Points())
    >>> IO.disconnect()

    Warnings
    --------
    Quite often (but unreliably!), running this function will get Matrix into a state in which it will return one/two
    lines (unless play/pause is clicked manually) when operating manually. To restore this functionality, run
    utils.utils.restore_z_functionality() and restart your scan.
    """
    global scan_dir_x, scan_dir_y, line_count_y, view_count, tot_packets, xydata

    # Setup and parsing parameters
    allowed = ["new", "pause", "continue"]
    if mode not in allowed:
        raise ValueError(f"Mode must be one of {allowed}")

    if num_lines == 'all':
        num_lines = mo.xy_scanner.Lines()

    x_dir_dict = {"Forward": "Fw",
                  "Backward": "Bw"}
    x_direction_strings = [x_dir_dict[x_dir] for x_dir in x_direction.split("-")]
    y_direction_strings = y_direction.split("-")

    # Set triggers
    if x_direction == "Forward-Backward":
        mo.xy_scanner.X_Retrace(True)
    else:
        mo.xy_scanner.X_Retrace(False)

    if y_direction == "Up-Down":
        if num_lines != 'all' and num_lines != mo.xy_scanner.Lines():  # Force set lines instead?
            raise ValueError("If scanning in both directions, num_lines cannot be set")
        mo.xy_scanner.Y_Retrace(True)
    else:
        mo.xy_scanner.Y_Retrace(False)

    # while not mo.xy_scanner.Enable_Scan():  # Enforce that we won't take data during relocation and confuse the CU
    #     print("waiting")
    #     sleep(0.1)

    # Set counters and pre-allocate
    view_count = [None, None]
    scan_dir_y = 0
    scan_dir_x = 1
    line_count_y = 0
    tot_packets = 0
    xydata = np.zeros((2, 2, num_lines, mo.xy_scanner.Points()))
    xydata[:] = np.nan

    def view_xy_callback():
        global scan_dir_x, scan_dir_y, line_count_y, view_count, tot_packets, xydata
        tot_packets += 1
        mo.tot_packets = tot_packets
        line_count_y += 1
        if x_direction == "Forward-Backward":
            scan_dir_x = int(not (bool(scan_dir_x)))
            line_count_y = line_count_y - scan_dir_x
        scan_dir_y = (line_count_y - 1) // num_lines

        data_size = mo.view.Data_Size()
        data_pts = np.array(mo.sample_data(data_size))

        xydata[scan_dir_y, int(not (bool(scan_dir_x))), (line_count_y - (num_lines * scan_dir_y)) - 1, :] = data_pts
        view_count = [mo.view.Run_Count(), mo.view.Cycle_Count()]

        # pbar.update(1)
        # # pbar.set_postfix({"Scanline": mo.view.Packet_Count()})
        # if tot_packets % 2 == 1:
        #     pbar.set_postfix({"Scanline": mo.view.Packet_Count()+1})

    # Enable channels
    for x_direction_string in x_direction_strings:
        IO.enable_channel(f"{channel_name}_{x_direction_string}")
        mo.view.Data(view_xy_callback)
        mo.allocate_sample_memory(mo.xy_scanner.Points())

    if mode == 'new':
        mo.experiment.start()
    elif mode == 'pause':
        mo.experiment.resume()
    else:
        pass

    # Get the data
    # pbar = tqdm(total=num_lines * len(x_direction_strings) * len(y_direction_strings))
    while tot_packets < num_lines * len(x_direction_strings) * len(y_direction_strings) \
            and mo.mate.rc == mo.mate.rcs['RMT_SUCCESS']:
        mo.wait_for_event()

    if mode == 'new':
        mo.experiment.stop()
    elif mode == 'pause':
        mo.experiment.pause()
    else:
        pass

    # Pretty the output to make physical sense
    xydata = np.flip(xydata, axis=2)
    if len(x_direction_strings) != 2:
        xydata = xydata[:, 0, :, :]
    if len(y_direction_strings) != 2:
        xydata = xydata[0, :, :, :]
    np.squeeze(xydata)

    # Return nicely
    if return_filename:
        filename = f"{mo.experiment.Result_File_Path()}\\{mo.experiment.Result_File_Name()}--{mo.view.Run_Count()}_{mo.view.Cycle_Count()}.Z_mtrx"
        return xydata, filename
    else:
        return xydata


if __name__ == "__main__":
    IO.connect()
    set_points_lines(128)
    xydata1 = get_xy_scan(channel_name="Z", x_direction="Forward", y_direction="Up", num_lines=1)

    plot_xy(xydata1, view_count, pixel_scale=mo.xy_scanner.Width() * 1e9 / mo.xy_scanner.Points())

    mo.experiment.Result_File_Name()



# Oliver Gordon, 2019

from nOmicron.mate import objects as mo
from nOmicron.utils import utils

def connect():
    """Connect to the Matrix. Matrix must be open and initalised."""

    print("Connecting to the Matrix...")
    mo.mate.testmode = False
    mo.mate.connect()
    utils.is_online()
    print("Connected Successfully!")


def disconnect():
    """Disconnect from the Matrix."""
    print("Disconnecting...")
    mo.experiment.stop()
    mo.mate.disconnect()
    print("Disconnected Successfully!")


def enable_channel(channel_name):
    """Enables a channel (e.g. Z_Fw, I_Bw, I_t, Aux1_t, I_V, I_Z) for measurement.
    Examples
    --------
    >>> from nOmicron.microscope import IO
    >>> IO.connect()
    >>> IO.enable_channel("I_t")
    """

    channel_name = channel_name.replace("(", "_").replace(")","")

    if channel_name[-1] == "w":
        utils.is_channel_real(channel_name[:-3])
    else:
        utils.is_channel_real(channel_name)

    if len(channel_name) != 1 and channel_name[-1] != "t" and channel_name[-1] != "w":
        channel_name += "_Spec"

    mo.channel_name = channel_name
    if channel_name[-1] == "t":
        mo.get_clock_name(mo.channel_name)

    mo.view.Deliver_Data(True)


def disable_channel():
    """Disables a channel from passing data."""
    if mo.channel_name == '':
        raise IOError("No channel to disable")

    mo.view.Deliver_Data(False)
    mo.channel_name = ''


def set_clock(sample_time, sample_points):
    """Sets up the clock.

    Parameters
    ----------
    sample_time : float
    sample_points : int

    Warnings
    --------
    Only enable this with data channels with a Clock (i.e. only continuous channels like I_t, df_t)

    Examples
    --------
    >>> from nOmicron.microscope import IO
    >>> IO.connect()
    >>> IO.enable_channel("Z_t")
    >>> IO.set_clock(1e-2, 200)  # 100 milliseconds, 200 points will be acquired on next trigger
    """
    if "Clock" not in mo.mate.deployment_parameter(mo.mate.scope, mo.channel_name, 'Trigger'):
        raise IOError("Attempted to set a clock for a non-clock object")

    mo.clock.Enable(False)
    mo.clock.Period(sample_time / sample_points)
    mo.clock.Samples(sample_points)
    mo.clock.Enable(True)


def intercept_target_position():
    """Intercepts the tip target position set with the mouse tool

    Returns
    -------
    target_pos : tuple

    Warnings
    --------
    If using this function to call a routine e.g. continuous_spectra.get_point_spectra(), do NOT execute a spectroscopy
    in Matrix to exectute this function - this will very often result in one data packet being missed. Instead, perform
    a non-trigger task, such as a tip relocation.

    Examples
    --------
    Intercept the xy position of a manual spectroscopy, and allow the data to be passed into python
    >>> from nOmicron.microscope.continuous_spectroscopy import get_point_spectra
    >>> pos = intercept_target_position()
    >>> v, I = get_point_spectra("I(V)", start_end=(0.5, -0.5), target_position=pos,
    >>>                   repeats=1, sample_points=100, sample_time=20e-3, forward_back=False)
    """
    current_pos = old_pos = mo.xy_scanner.Target_Position()
    while old_pos == current_pos:
        current_pos = mo.xy_scanner.Target_Position()

    return current_pos


def enable_pll():
    mo.pll.PLL_Enable(True)


def disable_pll():
    mo.pll.PLL_Enable(False)



#   Code up to mate4dummies 0.4.2 (at time of branch) is Copyright © 2015 - 2018 Stephan Zevenhuizen
#   MATE, (20-08-2018).
#   Additional changes by Oliver Gordon, 2019

import ctypes as _ctypes
import inspect as _inspect
import time as _time
from msvcrt import getch, kbhit
from random import random
from time import sleep

from nOmicron.utils.utils import is_parameter_allowable

from .mate import MATE as _MATE
from ..utils.errors import MatrixNotInitialisedError


class _Text(object):

    def __init__(self, text):
        self.text = text

    def AppendText(self, text):
        self.text += text


class _Channel(object):
    """Acquires and pre-processes measurement data."""

    def Enable(self, a=None, test=True):
        """Determines the state of the data acquisition facility. If True, the element will produce a data stream
        whenever it gets triggered.

        Warnings
        --------
        This parameter must not be changed if the channel is currently active, i.e. produces data
        """
        p = 'boolean'
        a = _process(p, [self, "Enable"], a, test)
        return a


class _Clock(object):
    """Generates a continuous clock signal at a variable frequency, for use as a trigger source for data acquisition.
     Data cannot be streamed live, only sent in selected buffer chunks."""

    def __init__(self, period):
        self.period = period

    def Enable(self, a=None, test=False):
        p = 'boolean'
        a = _process(p, [self, "Enable"], a, test)
        return a

    def Period(self, a=None):
        if mate.testmode and a:
            self.period = a
        test = self.period
        p = 'double'
        a = _process(p, [self, "Period"], a, test)
        return a

    def Samples(self, a=None, test=0):
        p = 'unsigned_integer'
        is_parameter_allowable(a, self.__class__.__name__, "Samples", test)
        a = _process(p, [self, "Samples"], a, test)
        return a


class _Experiment(object):
    """Core features of the experiment. Can read/change the state of the experiment."""

    def Bricklet_Ready(self, test=''):
        p = 'string'
        a = _process(p, [self, "Bricklet_Ready"], None, test)
        return a

    def Bricklet_Written(self, test=''):
        p = 'string'
        a = _process(p, [self, "Bricklet_Written"], None, test)
        return a

    def Name(self, test=''):
        p = 'string'
        a = _process(p, [self, "Name"], None, test)
        return a

    def Result_File_Name(self, test=''):
        p = 'string'
        a = _process(p, [self, "Result_File_Name"], None, test)
        return a

    def Result_File_Path(self, test=''):
        p = 'string'
        a = _process(p, [self, "Result_File_Path"], None, test)
        return a

    def State(self, test='test mode'):
        p = 'string'
        a = _process(p, [self, "State"], None, test)
        return a

    def pause(self):
        p = 'function'
        out = _process(p, [self, "pause"], None)

    def resume(self):
        p = 'function'
        out = _process(p, [self, "resume"], None)

    def start(self):
        p = 'function'
        out = _process(p, [self, "start"], None)

    def stop(self):
        p = 'function'
        out = _process(p, [self, "stop"], None)

    def restart(self):
        p = 'function'
        out = _process(p, [self, "restart"], None)

    def upload(self):
        p = 'function'
        out = _process(p, [self, "upload"], None)

    def unload(self):
        p = 'function'
        out = _process(p, [self, "unload"], None)


class _GapVoltageControl(object):
    """Control the gap voltage between the STM probe and the sample."""

    def Preamp_Range(self, a=None, test=0):
        p = 'enum'
        a = _process(p, [self, "Preamp_Range"], a, test)
        return a

    def Voltage(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Voltage", test)
        a = _process(p, [self, "Voltage"], a, test)
        return a

    def Tip_Cond_Enable_Feedback_Loop(self, a=None, test=True):
        p = 'boolean'
        a = _process(p, [self, "Tip_Cond_Enable_Feedback_Loop"], a, test)
        return a

    def Tip_Cond_Pulse_Time(self, a=None, test=0.0):
        """Tip pulse time."""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Tip_Cond_Pulse_Time", test)
        a = _process(p, [self, "Tip_Cond_Pulse_Time"], a, test)
        return a

    def Tip_Cond_Pulse_Voltage(self, a=None, test=0.0):
        """Tip pulse voltage."""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Tip_Cond_Pulse_Voltage", test)
        a = _process(p, [self, "Tip_Cond_Pulse_Voltage"], a, test)
        return a

    def Tip_Cond_Pulse_Preamp_Range(self, a=None, test=0.0):
        """Tip pulse voltage range. 0 -> (-1,1). 1-> (-10, 10)."""
        p = 'enum'
        # is_parameter_allowable(a, self.__class__.__name__, "ASDasd", test)
        a = _process(p, [self, "Tip_Cond_Pulse_Preamp_Range"], a, test)
        return a

    # def Tip_Cond_Pulse_Apply(self, f=None, *args, **kwargs):
    #     """Does a tip conditioning"""
    #     p = 'set_observed'
    #     out = _process(p, [self, "ASDASDASDASD"], f, *args, **kwargs)

    def Tip_Cond_Pulse_Apply(self):
        """Super powerful. Do something, then run the trigger function on completion"""
        p = 'function'
        out = _process(p, [self, "Tip_Cond_Pulse_Apply"], None)

    def Execute(self):
        """Execute spectroscopy"""
        p = 'function'
        out = _process(p, [self, "Execute"], None)


class _PiezoControl(object):
    """Controls tip approach."""

    def Approach(self, a=None, test=0):
        p = 'unsigned_integer'
        is_parameter_allowable(a, self.__class__.__name__, "Approach", test)
        a = _process(p, [self, "Approach"], a, test)
        return a

    def Move_Auto(self, a=None, test=True):
        p = 'boolean'
        a = _process(p, [self, "Move_Auto"], a, test)
        return a

    def Move_Backward(self, a=None, test=True):
        p = 'boolean'
        a = _process(p, [self, "Move_Backward"], a, test)
        return a

    def Move_Forward(self, a=None, test=True):
        p = 'boolean'
        a = _process(p, [self, "Move_Forward"], a, test)
        return a

    def Move_Tip_X_Minus(self, a=None, test=0):
        p = 'unsigned_integer'
        is_parameter_allowable(a, self.__class__.__name__, "Move_Tip_X_Minus", test)
        a = _process(p, [self, "Move_Tip_X_Minus"], a, test)
        return a

    def Move_Tip_X_Plus(self, a=None, test=0):
        p = 'unsigned_integer'
        is_parameter_allowable(a, self.__class__.__name__, "Move_Tip_X_Plus", test)
        a = _process(p, [self, "Move_Tip_X_Plus"], a, test)
        return a

    def Move_Tip_Y_Minus(self, a=None, test=0):
        p = 'unsigned_integer'
        is_parameter_allowable(a, self.__class__.__name__, "Move_Tip_Y_Minus", test)
        a = _process(p, [self, "Move_Tip_Y_Minus"], a, test)
        return a

    def Move_Tip_Y_Plus(self, a=None, test=0):
        p = 'unsigned_integer'
        is_parameter_allowable(a, self.__class__.__name__, "Move_Tip_Y_Plus", test)
        a = _process(p, [self, "Move_Tip_Y_Plus"], a, test)
        return a

    def Retract(self, a=None, test=0):
        p = 'unsigned_integer'
        is_parameter_allowable(a, self.__class__.__name__, "Retract", test)
        a = _process(p, [self, "Retract"], a, test)
        return a

    def isLocked(self, s, test=0):
        p = 'function'
        out = _process(p, [self, "isLocked"], s, test)
        return out


class _Regulator(object):
    """Controls the Z position of the probe."""

    def Enable_Z_Offset_Slew_Rate(self, a=None, test=False):
        p = 'boolean'
        a = _process(p, [self, "Enable_Z_Offset_Slew_Rate"], a, test)
        return a

    def Z_Ramp(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Z_Ramp", test)
        a = _process(p, [self, "Z_Ramp"], a, test)
        return a

    def Z_Ramp_Delay(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Z_Ramp_Delay", test)
        a = _process(p, [self, "Z_Ramp_Delay"], a, test)
        return a

    def Enable_Z_Ramp_Slew_Rate(self, a=None, test=False):
        p = 'boolean'
        a = _process(p, [self, "Enable_Z_Ramp_Slew_Rate"], a, test)
        return a

    def Z_Ramp_Slew_Rate(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Z_Ramp_Slew_Rate", test)
        a = _process(p, [self, "Z_Ramp_Slew_Rate"], a, test)
        return a

    def Feedback_Loop_Enabled(self, a=None, test=False):
        p = 'boolean'
        a = _process(p, [self, "Feedback_Loop_Enabled"], a, test)
        return a

    def Loop_Gain_1_I(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Loop_Gain_1_I", test)
        a = _process(p, [self, "Loop_Gain_1_I"], a, test)
        return a

    def Loop_Gain_2_I(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Loop_Gain_2_I", test)
        a = _process(p, [self, "Loop_Gain_2_I"], a, test)
        return a

    def Preamp_Range_1(self, a=None, test=0):
        p = 'enum'
        # is_parameter_allowable(a, self.__class__.__name__, "ASDasd", test)
        a = _process(p, [self, "Preamp_Range_1"], a, test)
        return a

    def Preamp_Range_2(self, a=None, test=0):
        p = 'enum'
        # is_parameter_allowable(a, self.__class__.__name__, "ASDasd", test)
        a = _process(p, [self, "Preamp_Range_2"], a, test)
        return a

    def Setpoint_1(self, a=None, test=0.0):
        p = 'double'
        # is_parameter_allowable(a, self.__class__.__name__, "ASDasd", test)
        a = _process(p, [self, "Setpoint_1"], a, test)
        return a

    def Setpoint_2(self, a=None, test=0.0):
        p = 'double'
        # is_parameter_allowable(a, self.__class__.__name__, "ASDasd", test)
        a = _process(p, [self, "Setpoint_2"], a, test)
        return a

    def Z_Offset(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Z_Offset", test)
        a = _process(p, [self, "Z_Offset"], a, test)
        return a

    def Z_Offset_Slew_Rate(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Z_Offset_Slew_Rate", test)
        a = _process(p, [self, "Z_Offset_Slew_Rate"], a, test)
        return a

    def Z_Out(self, test=0.0):
        p = 'double'
        a = _process(p, [self, "Z_Out"], None, test)
        return a

    def Setpoint_Detected(self, a=None, test=False):
        p = 'boolean'
        a = _process(p, [self, "Setpoint_Detected"], a, test)
        return a


class _View(object):
    """Allows for the delivery of data. Open the channel with IO.py, set Deliver_Data to True and return data
    with Data() and a callback if desired."""

    def Data(self, f=None, *args, **kwargs):
        p = 'set_observed'
        out = _process(p, [self, "Data"], f, *args, **kwargs)

    def Data_Size(self, test=None):
        if mate.testmode and not test:
            test = _p_values[0].values[0].realArray[0][0].length
        p = 'unsigned_integer'
        a = _process(p, [self, "Data_Size"], None, test)
        return a

    def Deliver_Data(self, a=None, test=False):
        p = 'boolean'
        a = _process(p, [self, "Deliver_Data"], a, test)
        return a

    def Run_Count(self, test=0):
        p = 'unsigned_integer'
        a = _process(p, [self, "Run_Count"], None, test)
        return a

    def Cycle_Count(self, test=0):
        p = 'unsigned_integer'
        a = _process(p, [self, "Cycle_Count"], None, test)
        return a

    def Packet_Count(self, test=0):
        p = 'unsigned_integer'
        a = _process(p, [self, "Packet_Count"], None, test)
        return a


class _XYScanner(object):
    """Controls general scan parameters, tip relocation, and triggering of functions after tip relocation"""

    def Angle(self, a=None, test=0):
        """Integer describing angle of scan"""
        p = 'integer'
        is_parameter_allowable(a, self.__class__.__name__, "Angle", test)
        a = _process(p, [self, "Angle"], a, test)
        return a

    def Area(self, a=None, test=[0.0, 0.0]):
        """Scan area in metres"""
        p = 'pair'
        is_parameter_allowable(a, self.__class__.__name__, "Area", test)
        a = _process(p, [self, "Area"], a, test)
        return a

    def Execute_Port_Colour(self, a=None, test=0):
        """Here be dragons."""
        p = 'string'
        a = _process(p, [self, "Execute_Port_Colour"], a, test)
        return a

    def Points(self, a=None, test=0):
        """Integer number of points to scan with. Fixed to Lines unless Points_Lines_Constrained = False"""
        p = 'unsigned_integer'
        is_parameter_allowable(a, self.__class__.__name__, "Points", test)
        # print(self, "ASDasd")
        a = _process(p, [self, "Points"], a, test)
        return a

    def Lines(self, a=None, test=0):
        """Integer number of lines to scan per image. Fixed to Points unless Points_Lines_Constrained = False"""
        p = 'unsigned_integer'
        is_parameter_allowable(a, self.__class__.__name__, "Lines", test)
        a = _process(p, [self, "Lines"], a, test)
        return a

    def Points_Lines_Constrained(self, a=None, test=False):
        """Boolean of if scan points and lines should be locked together"""
        p = 'boolean'
        a = _process(p, "Points_Lines_Constrained", a, test)
        return a

    def Enable_Scan(self, a=None, test=False):
        """Enable/Disable the scan"""
        p = 'boolean'
        a = _process(p, [self, "Enable_Scan"], a, test)
        return a

    def Offset(self, a=None, test=[0.0, 0.0]):
        """Scan-sample offset, in metres. Identical to setting X_Offset and Y_Offset separately."""
        p = 'pair'
        is_parameter_allowable(a, self.__class__.__name__, "Offset", test)
        a = _process(p, [self, "Offset"], a, test)
        return a

    def Enable_Drift_Compensation(self, a=None, test=False):
        """Boolean to enable drift compensation."""
        p = 'boolean'
        a = _process(p, [self, "Enable_Drift_Compensation"], a, test)
        return a

    def Enable_Plane_Slope(self, a=None, test=False):
        """Boolean for sample tilt correction. See also Plane_X_Slope, Plane_Y_Slope and Detect_Slope."""
        p = 'boolean'
        a = _process(p, [self, "Enable_Plane_Slope"], a, test)
        return a

    def Plane_X_Slope(self, a=None, test=0.0):
        """Sample slope in X direction. Percentage in range(-100, 100)"""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Plane_X_Slope", test)
        a = _process(p, [self, "Plane_X_Slope"], a, test)
        return a

    def Plane_Y_Slope(self, a=None, test=0.0):
        """Sample slope in Y direction. Percentage in range(-100, 100)"""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Plane_Y_Slope", test)
        a = _process(p, [self, "Plane_Y_Slope"], a, test)
        return a

    def Move_Raster_Time_Constrained(self, a=None, test=False):
        """Boolean of if scan raster and move should be locked together"""
        p = 'boolean'
        a = _process(p, [self, "Move_Raster_Time_Constrained"], a, test)
        return a

    def Raster_Time(self, a=None, test=0.0):
        """Time in seconds to move between scanning points. Identical to XYScanner.Move_Raster_Time. Real value is
        shown with Scan_Speed"""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Raster_Time", test)
        a = _process(p, [self, "Raster_Time"], a, test)
        return a

    def Move_Raster_Time(self, a=None, test=0.0):
        """Time in seconds to move between scanning points. Identical to XYScanner.Move_Raster_Time. Real value is
        shown with Scan_Speed"""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Move_Raster_Time", test)
        a = _process(p, [self, "Move_Raster_Time"], a, test)
        return a

    def Speed_Adaption(self, a=None, test=0):
        """Pick the speed adaption mode. 0 = constant line freq., 1 = constant scan speed"""
        p = 'enum'
        a = _process(p, [self, "Speed_Adaption"], a, test)
        return a

    def Scan_Speed(self, a=None, test=0.0):
        """Time in seconds to move between scanning points. Identical to XYScanner.Move_Raster_Time. Real value is
        shown with Scan_Speed"""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Scan_Speed", test)
        a = _process(p, [self, "Scan_Speed"], a, test)
        return a

    def Enable_Execute_Port(self, a=None, test=False):
        """Set to True to return the probe to the position set by XYScanner.Store_Current_Position"""
        p = 'boolean'
        a = _process(p, [self, "Enable_Execute_Port"], a, test)
        return a

    def Return_To_Stored_Position(self, a=None, test=False):
        """Set to True to return the probe to the position set by XYScanner.Store_Current_Position"""
        p = 'boolean'
        a = _process(p, [self, "Return_To_Stored_Position"], a, test)
        return a

    def Store_Current_Position(self, a=None, test=False):
        """Set to True to store the current scan position in hardware, to later recall with
        XYScanner.Return_To_Stored_Position"""
        p = 'boolean'
        a = _process(p, [self, "Store_Current_Position"], a, test)
        return a

    def Target_Position(self, a=None, test=[0.0, 0.0]):
        """Target position for tip within scanning window. NOT in metres, but co-ordinates in range (-1, 1)"""
        p = 'pair'
        a = _process(p, [self, "Target_Position"], a, test)
        return a

    def Trigger_Execute_At_Target_Position(self, a=None, test=False):
        """Boolean of if some function (e.g. spectroscopy) should be executed when the tip reaches the position
        described by XYScanner.Target_Position([x, y]), and beginning the movement with XYScanner.move()"""
        p = 'boolean'
        a = _process(p, [self, "Trigger_Execute_At_Target_Position"], a, test)
        return a

    def Width_Height_Constrained(self, a=None, test=False):
        """Boolean of if scan width and height should be locked together"""
        p = 'boolean'
        a = _process(p, [self, "Width_Height_Constrained"], a, test)
        return a

    def Width(self, a=None, test=0.0):
        """Scan width in metres"""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Width", test)
        a = _process(p, [self, "Width"], a, test)
        return a

    def Height(self, a=None, test=0.0):
        """Scan height in metres"""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Height", test)
        a = _process(p, [self, "Height"], a, test)
        return a

    def XY_Position_Report(self, test=[0.0, 0.0]):
        p = 'pair'
        a = _process(p, [self, "XY_Position_Report"], None, test)
        return a

    def X_Offset(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "X_Offset", test)
        a = _process(p, [self, "X_Offset"], a, test)
        return a

    def X_Drift(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "X_Drift", test)
        a = _process(p, [self, "X_Drift"], a, test)
        return a

    def X_Retrace(self, a=None, test=False):
        p = 'boolean'
        a = _process(p, [self, "X_Retrace"], a, test)
        return a

    def X_Retrace_Done(self, f=None, *args, **kwargs):
        p = 'set_observed'
        out = _process(p, [self, "X_Retrace_Done"], f, *args, **kwargs)

    def X_Retrace_Trigger(self, a=None, test=False):
        p = 'boolean'
        a = _process(p, [self, "X_Retrace_Trigger"], a, test)
        return a

    def X_Trace_Done(self, f=None, *args, **kwargs):
        p = 'set_observed'
        out = _process(p, [self, "X_Trace_Done"], f, *args, **kwargs)

    def X_Trace_Trigger(self, a=None, test=False):
        p = 'boolean'
        a = _process(p, [self, "X_Trace_Trigger"], a, test)
        return a

    def Y_Drift(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Y_Drift", test)
        a = _process(p, [self, "Y_Drift"], a, test)
        return a

    def Y_Offset(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Y_Offset", test)
        a = _process(p, [self, "Y_Offset"], a, test)
        return a

    def Y_Retrace(self, a=None, test=False):
        """If we should trigger and do something (e.g. acquire data) after a complete upwards (False) or upwards AND
        downwards (True) scan"""
        p = 'boolean'
        a = _process(p, [self, "Y_Retrace"], a, test)
        return a

    def Y_Retrace_Done(self, f=None, *args, **kwargs):
        """Allows us to call a callback (e.g. for data acquisition) after completing an entire forward AND backward
        scan. Y_Retrace_Trigger must first be set to True."""
        p = 'set_observed'
        out = _process(p, [self, "Y_Retrace_Done"], f, *args, **kwargs)

    def Y_Retrace_Trigger(self, a=None, test=False):
        """If we want to trigger and do something (e.g. acquire data) after completing an entire
        forward AND backward scan"""
        p = 'boolean'
        a = _process(p, [self, "Y_Retrace_Trigger"], a, test)
        return a

    def Y_Trace_Done(self, f=None, *args, **kwargs):
        """Allows us to call a callback (e.g. for data acquisition) after completing an entire forward scan.
        Y_Trace_Trigger must first be set to True."""
        p = 'set_observed'
        out = _process(p, [self, "Y_Trace_Done"], f, *args, **kwargs)

    def Y_Trace_Trigger(self, a=None, test=False):
        """If we want to trigger and do something (e.g. acquire data) after completing an entire forward scan"""
        p = 'boolean'
        a = _process(p, [self, "Y_Trace_Trigger"], a, test)
        return a

    def move(self):
        """Move the tip to the position set with Target_Position"""
        p = 'function'
        out = _process(p, [self, "move"], None)

    def resume(self):
        """Resume the imaging scan"""
        p = 'function'
        out = _process(p, [self, "resume"], None)

    def execute(self):
        """Super powerful. Do something, then run the trigger function on completion"""
        p = 'function'
        out = _process(p, [self, "execute"], None)

    def Operation_Cancelled(self):
        """Move the tip to the position set with Target_Position"""
        p = 'function'
        out = _process(p, [self, "Operation_Cancelled"], None)


class _Spectroscopy:
    """Controls spectroscopy. Device 1&2 are the first two channels in the Matrix window (?)."""

    def Enable_Feedback_Loop(self, a=None, test=False):
        """Enable the feedback loop"""
        p = 'boolean'
        a = _process(p, [self, "Enable_Feedback_Loop"], a, test)
        return a

    def Disable_Feedback_Loop(self, a=None, test=False):
        """Disable the feedback loop"""
        p = 'boolean'
        a = _process(p, [self, "Disable_Feedback_Loop"], a, test)
        return a

    def Spectroscopy_Mode(self, a=None, test=0):
        """Pick the channel to perform the spectroscopy on when calling Spectroscopy.execute().
         0->channel 1, 1->channel 2, 2->dragons"""
        p = 'enum'
        a = _process(p, [self, "Spectroscopy_Mode"], a, test)
        return a

    def execute(self):
        """Execute spectroscopy"""
        p = 'function'
        out = _process(p, [self, "execute"], None)

    def Device_1_Repetitions(self, a=None, test=0):
        """Integer describing number of repeats of spectra on channel 1"""
        p = 'unsigned_integer'
        is_parameter_allowable(a, self.__class__.__name__, "Device_1_Repetitions", test)
        a = _process(p, [self, "Device_1_Repetitions"], a, test)
        return a

    def Device_1_Start(self, a=None, test=0.0):
        """Spectroscopy start value on channel 1"""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Device_1_Start", test)
        a = _process(p, [self, "Device_1_Start"], a, test)
        return a

    def Device_1_End(self, a=None, test=0.0):
        """Spectroscopy end value on channel 1"""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Device_1_End", test)
        a = _process(p, [self, "Device_1_End"], a, test)
        return a

    def Device_1_Points(self, a=None, test=0):
        """Spectroscopy points on channel 1"""
        p = 'unsigned_integer'
        is_parameter_allowable(a, self.__class__.__name__, "Device_1_Points", test)
        a = _process(p, [self, "Device_1_Points"], a, test)
        return a

    def Raster_Time_1(self, a=None, test=0):
        """Time in seconds to acquire data at each spectroscopy step on channel 1."""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Raster_Time_1", test)
        a = _process(p, [self, "Raster_Time_1"], a, test)
        return a

    def Device_1_Offset_Delay(self, a=None, test=0.0):
        """Offset time"""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Device_1_Offset_Delay", test)
        a = _process(p, [self, "Device_1_Offset_Delay"], a, test)
        return a

    def Enable_Device_1_Ramp_Reversal(self, a=None, test=False):
        """More dragons."""
        p = 'boolean'
        a = _process(p, [self, "Enable_Device_1_Ramp_Reversal"], a, test)
        return a

    def Device_2_Repetitions(self, a=None, test=0):
        """Integer describing number of repeats of spectra on channel 2"""
        p = 'unsigned_integer'
        is_parameter_allowable(a, self.__class__.__name__, "Device_2_Repetitions", test)
        a = _process(p, [self, "Device_2_Repetitions"], a, test)
        return a

    def Device_2_Start(self, a=None, test=0.0):
        """Spectroscopy start value on channel 2"""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Device_2_Start", test)
        a = _process(p, [self, "Device_2_Start"], a, test)
        return a

    def Device_2_End(self, a=None, test=0.0):
        """Spectroscopy end value on channel 2"""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Device_2_End", test)
        a = _process(p, [self, "Device_2_End"], a, test)
        return a

    def Device_2_Points(self, a=None, test=0):
        """Spectroscopy points on channel 2"""
        p = 'unsigned_integer'
        is_parameter_allowable(a, self.__class__.__name__, "Device_2_Points", test)
        a = _process(p, [self, "ASDASDASDASD"], a, test)
        return a

    def Raster_Time_2(self, a=None, test=0):
        """Time in seconds to acquire data at each spectroscopy step on channel 2."""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Raster_Time_2", test)
        a = _process(p, [self, "Raster_Time_2"], a, test)
        return a

    def Device_2_Offset_Delay(self, a=None, test=0.0):
        """Offset time"""
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Device_2_Offset_Delay", test)
        a = _process(p, [self, "Device_2_Offset_Delay"], a, test)
        return a

    def Enable_Device_2_Ramp_Reversal(self, a=None, test=False):
        """More dragons."""
        p = 'boolean'
        a = _process(p, [self, "Enable_Device_2_Ramp_Reversal"], a, test)
        return a


class _PLLControl(object):
    """Controls the PLL regulator"""

    def Enable_Tip_Protection(self, a=None, test=False):
        is_parameter_allowable(a, self.__class__.__name__, "Enable_Tip_Protection", test)
        p = 'boolean'
        a = _process(p, [self, "Enable_Tip_Protection"], a, test)
        return a

    def Enable_Constant_Excitation_Mode(self, a=None, test=False):
        is_parameter_allowable(a, self.__class__.__name__, "Enable_Constant_Excitation_Mode", test)
        p = 'boolean'
        a = _process(p, [self, "Enable_Constant_Excitation_Mode"], a, test)
        return a

    def PLL_Enable(self, a=None, test=False):
        is_parameter_allowable(a, self.__class__.__name__, "PLL_Enable", test)
        p = 'boolean'
        a = _process(p, [self, "PLL_Enable"], a, test)
        return a

    def Auto_Phase(self, a=None, test=False):
        is_parameter_allowable(a, self.__class__.__name__, "Auto_Phase", test)
        p = 'boolean'
        a = _process(p, [self, "Auto_Phase"], a, test)
        return a

    def PLL_Centre_Frequency(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "PLL_Centre_Frequency", test)
        a = _process(p, [self, "PLL_Centre_Frequency"], a, test)
        return a

    def Amplitude_Detection_Mode(self, a=None, test=0):
        """0 = frequency selective, 1 = broadband"""
        p = 'enum'
        # is_parameter_allowable(a, self.__class__.__name__, "ASDasd", test)
        a = _process(p, [self, "Amplitude_Detection_Mode"], a, test)
        return a

    def Sensor_Resonance_Frequency_Range(self, a=None, test=0):
        """0 = 12.5-40 kHz, 1 = 15-60 kHz, 2 = 50-200 kHz, 3 = 150-500 kHz, 4 = 375-1250 kHz"""
        p = 'enum'
        # is_parameter_allowable(a, self.__class__.__name__, "ASDasd", test)
        a = _process(p, [self, "Sensor_Resonance_Frequency_Range"], a, test)
        return a

    def Non_Contact_Mode(self, a=None, test=0):
        """0 = constant amplitude, 1 = constant excitation, 2 = self excitation"""
        p = 'enum'
        # is_parameter_allowable(a, self.__class__.__name__, "ASDasd", test)
        a = _process(p, [self, "Non_Contact_Mode"], a, test)
        return a

    def Excitation_Attenuation_Constant_Amplitude(self, a=None, test=0):
        """0 = no attenuation, 1 = 0.1, 2 = 0.01, 3 = 0.001"""
        p = 'enum'
        # is_parameter_allowable(a, self.__class__.__name__, "ASDasd", test)
        a = _process(p, [self, "Excitation_Attenuation_Constant_Amplitude"], a, test)
        return a

    def Excitation_Attenuation_Constant_Excitation(self, a=None, test=0):
        """0 = no attenuation, 1 = 0.1, 2 = 0.01, 3 = 0.001"""
        p = 'enum'
        # is_parameter_allowable(a, self.__class__.__name__, "ASDasd", test)
        a = _process(p, [self, "Excitation_Attenuation_Constant_Excitation"], a, test)
        return a

    def Excitation_Attenuation_Self_Excitation(self, a=None, test=0):
        """0 = no attenuation, 1 = 0.1, 2 = 0.01, 3 = 0.001"""
        p = 'enum'
        # is_parameter_allowable(a, self.__class__.__name__, "ASDasd", test)
        a = _process(p, [self, "Excitation_Attenuation_Self_Excitation"], a, test)
        return a

    def Amplitude_Loop_Gain_I_Constant_Amplitude(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Amplitude_Loop_Gain_I_Constant_Amplitude", test)
        a = _process(p, [self, "Amplitude_Loop_Gain_I_Constant_Amplitude"], a, test)
        return a

    def Amplitude_Loop_Gain_P_Constant_Amplitude(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Amplitude_Loop_Gain_P_Constant_Amplitude", test)
        a = _process(p, [self, "Amplitude_Loop_Gain_P_Constant_Amplitude"], a, test)
        return a

    def Amplitude_Loop_Gain_I_Self_Excitation(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Amplitude_Loop_Gain_I_Self_Excitation", test)
        a = _process(p, [self, "Amplitude_Loop_Gain_I_Self_Excitation"], a, test)
        return a

    def Amplitude_Loop_Gain_P_Self_Excitation(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "Amplitude_Loop_Gain_P_Self_Excitation", test)
        a = _process(p, [self, "Amplitude_Loop_Gain_P_Self_Excitation"], a, test)
        return a

    def PLL_Loop_Gain_I(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "PLL_Loop_Gain_I", test)
        a = _process(p, [self, "PLL_Loop_Gain_I"], a, test)
        return a

    def PLL_Loop_Gain_P(self, a=None, test=0.0):
        p = 'double'
        is_parameter_allowable(a, self.__class__.__name__, "PLL_Loop_Gain_P", test)
        a = _process(p, [self, "PLL_Loop_Gain_P"], a, test)
        return a

    def Amplitude(self, test=0.0):
        """PLL amplitude. Read only"""
        p = 'double'
        a = _process(p, [self, "Amplitude"], None, test)
        return a

    def Delta_f(self, test=0.0):
        """PLL delta_f (Hz). Read only"""
        p = 'double'
        a = _process(p, [self, "Delta_f"], None, test)
        return a

    def Sensor_Frequency(self, test=0.0):
        """PLL sensor frequency (Hz). Read only"""
        p = 'double'
        a = _process(p, [self, "Sensor_Frequency"], None, test)
        return a

    def Damping(self, test=0.0):
        """PLL damping voltage (V). Read only"""
        p = 'double'
        a = _process(p, [self, "Damping"], None, test)
        return a

    def PLL_Locked(self, test=False):
        """See if the phase has been locked. Read only"""
        p = 'boolean'
        a = _process(p, [self, "PLL_Locked"], None, test)
        return a


def _process(p, caller, a, *args, **kwargs):
    global event_objects, _test_event_object
    if not hasattr(mate, "lib_mate"):
        raise MatrixNotInitialisedError()

    if caller[0].__class__.__name__ == '_Experiment':
        obj = mate.scope + '.' + caller[1]
    elif caller[0].__class__.__name__ == '_Channel':
        obj = channel_name + '.' + caller[1]
    elif caller[0].__class__.__name__ == '_Clock':
        obj = clock_name + '.' + caller[1]
    elif caller[0].__class__.__name__ == '_View':
        obj = 'View.' + channel_name + '.' + caller[1]
    else:
        obj = caller[0].__class__.__name__[1:] + '.' + caller[1]
    if p == 'string':
        if isinstance(a, str):
            func_params = [None, 'setString', obj, a]
            _, mate.rc = mate.remote_access(func_params, mate.rc)
        func_params = [args[0], 'getString', obj]
        a, mate.rc = mate.remote_access(func_params, mate.rc)
    elif p == 'boolean':
        if isinstance(a, bool):
            func_params = [None, 'setBoolean', obj, a]
            _, mate.rc = mate.remote_access(func_params, mate.rc)
        func_params = [args[0], 'getBoolean', obj]
        a, mate.rc = mate.remote_access(func_params, mate.rc)
    elif p == 'integer':
        if isinstance(a, int):
            func_params = [None, 'setInteger', obj, a]
            _, mate.rc = mate.remote_access(func_params, mate.rc)
        func_params = [args[0], 'getInteger', obj]
        a, mate.rc = mate.remote_access(func_params, mate.rc)
    elif p == 'unsigned_integer':
        if isinstance(a, int):
            func_params = [None, 'setInteger', obj, a]
            _, mate.rc = mate.remote_access(func_params, mate.rc)
        func_params = [args[0], 'getInteger', obj, _ctypes.c_uint()]
        a, mate.rc = mate.remote_access(func_params, mate.rc)
    elif p == 'enum':
        if isinstance(a, int):
            func_params = [None, 'setEnum', obj, a]
            _, mate.rc = mate.remote_access(func_params, mate.rc)
        func_params = [args[0], 'getEnum', obj]
        a, mate.rc = mate.remote_access(func_params, mate.rc)
    elif p == 'double':
        if isinstance(a, (int, float)):
            func_params = [None, 'setDouble', obj, a]
            _, mate.rc = mate.remote_access(func_params, mate.rc)
        func_params = [args[0], 'getDouble', obj]
        a, mate.rc = mate.remote_access(func_params, mate.rc)
    elif p == 'pair':
        if (isinstance(a, (list, tuple)) and len(a) == 2 and
                isinstance(a[0], (int, float)) and
                isinstance(a[1], (int, float))):
            func_params = [None, 'setPair', obj, a[0], a[1]]
            _, mate.rc = mate.remote_access(func_params, mate.rc)
        func_params = [args[0], 'getPair', obj]
        a, mate.rc = mate.remote_access(func_params, mate.rc)
    elif p == 'function':
        flat_value = mate.flat_values(0, 0, 1).values[0]
        if a and mate.testmode:
            if isinstance(args[0], bool):
                flat_value.type = 4
                flat_value.boolean = args[0]
            elif isinstance(args[0], int):
                flat_value.type = 1
                flat_value.integer = args[0]
            elif isinstance(args[0], float):
                flat_value.type = 2
                flat_value.real = args[0]
            elif (isinstance(args[0], (list, tuple)) and len(args[0]) == 2 and
                  isinstance(args[0][0], (int, float)) and
                  isinstance(args[0][1], (int, float))):
                flat_value.type = 6
                flat_value.pairX = args[0][0]
                flat_value.pairY = args[0][1]
        p_args = _ctypes.pointer(mate.flat_values(255, 0, 1))
        if a:
            if not isinstance(a, str):
                a = ''
            a = a.encode()
            p_args[0].values[0].type = mate.ValueType.vt_STRING
            p_args[0].values[0].string[0][0].length = len(a)
            p_args[0].values[0].string[0][0].text = a
        func_params = [flat_value, p, obj, p_args]
        out, mate.rc = mate.remote_access(func_params, mate.rc)
        if a:
            if out.type == 1:
                a = out.integer
            elif out.type == 2:
                a = out.real
            elif out.type == 4:
                a = bool(out.boolean)
            elif out.type == 5:
                a = out.enumeration
            elif out.type == 6:
                a = out.pairX, out.pairY
            else:
                a = None
    elif p == 'set_observed':
        if mate.testmode:
            _test_event_object = mate.scope + '::' + obj
        event_objects[obj] = [a, args, kwargs]
        func_params = [None, 'setObserved', obj, int(hasattr(a, '__call__'))]
        _, mate.rc = mate.remote_access(func_params, mate.rc)
    else:
        mate.rc = 0
    _check_rc()
    return a


def _no_event():
    global event_out, esc
    func_params = [(_test_event_object, 0, _p_values), 'getEvent']
    event_out, mate.rc = mate.remote_access(func_params, mate.rc)
    no_event = mate.rc == mate.rcs['RMT_NOEVENT']
    if mate.rc == mate.rcs['RMT_SUCCESS'] or mate.testmode:
        v = event_objects[event_out[0][len(mate.scope) + 2:]]
        v[0](*v[1], **v[2])
    # if kbhit():
    #     esc = ord(getch()) == 27
    return no_event# and not esc


def _check_rc():
    if mate.online and not mate.testmode:
        mate.exit_handler(mate.rc)


def _exit_handler():
    pass


def wait_for_event():
    log.AppendText('Waiting for event...\n')
    while _no_event():
        pass
        # sleep(0.01)


def get_clock_name(channel_name):
    global clock_name
    clock_name = mate.deployment_parameter(mate.scope, channel_name, 'Trigger')


def allocate_sample_memory(samples, test=None):
    global _p_values
    _p_values = _ctypes.pointer(mate.flat_values(0, samples, 1))
    if mate.testmode:
        if not test:
            b = 0.001
            a = -0.001
            test = [(b - a) * random() + a for i in range(samples)]
        _p_values[0].values[0].realArray[0][0].values = (_ctypes.c_double *
                                                         len(test))(*test)


def sample_data(data_size):
    return event_out[2][0].values[0].realArray[0][0].values[:data_size]


log = _Text('Starting log on ' + _time.strftime('%A, %d %B %Y %H:%M:%S',
                                                _time.localtime()) + '.\n')
mate = _MATE(log, _exit_handler, False)
event_objects = {}
_test_event_object = ''
_p_values = _ctypes.pointer(mate.flat_values(0, 0, 1))
event_out = ('', 0, _p_values)
channel_name = ''
clock_name = ''
channel = _Channel()
clock = _Clock(1.0)
experiment = _Experiment()
gap_voltage_control = _GapVoltageControl()
piezo_control = _PiezoControl()
pll = _PLLControl()
regulator = _Regulator()
view = _View()
xy_scanner = _XYScanner()
spectroscopy = _Spectroscopy()
tot_packets = 0
esc = False



# Oliver Gordon, 2019

import inspect
import json
import sys
import warnings

from bs4 import BeautifulSoup

from nOmicron.mate import objects as mo
from nOmicron.utils.errors import MatrixNotInitialisedError, MatrixParameterOutOfRangeWarning, \
    MatrixUnsupportedOperationError, MatrixInvalidDataTypeError


def is_online():
    """Tests if matrix is connected and running."""
    if not mo.mate.online or mo.experiment.Result_File_Name() == 'void':
        raise MatrixNotInitialisedError()
    return mo.mate.online


def is_channel_real(channel_name):
    """Tests if chosen channel exists and is enabled.

    Parameters
    ----------
    channel_name : str
    channel_name : str
        The name of the channel
    """

    is_online()
    response = mo.mate.deployment_parameter(mo.mate.scope, channel_name, 'Trigger') == ''
    if response:
        raise MatrixUnsupportedOperationError(f"Requested channel '{channel_name}' is not available in this \n"
                                              f"experiment. Available channels = {get_allowed_channels()}")
    return not response


def is_data_size_set():
    response = mo.view.Data_Size() is int
    if not response:
        warnings.warn("Data size has not been set/acknowledged by Matrix. No data will be sent!")

    return response


def is_parameter_within_allowed_range(parameter_name, parameter_value, config_file="control_limits.json"):
    # TODO: Include the minmax thing, and update the json file to match!
    data = json.loads(open(config_file).read())
    response = data[parameter_name][0] <= parameter_value <= data[parameter_name][1]

    if not response:
        raise ValueError(
            f"{parameter_name} is not allowed (requested {parameter_value}, should be within {data[parameter_name]})")

    return response


def _force_set_scope(scope_name):
    """Forcibly sets the experiment scope e.g. STM_Basic"""
    mo.mate.scope = scope_name
    mo.mate.lib_mate.setScopeName(scope_name)


def _friendly_name_to_mate(module_name):
    """Slightly hacky code to convert friendly mo.objects names to the mate4dummies mo.mate format"""
    all_modules = {"channel": "_Channel",
                   "clock": "_Clock",
                   "experiment": "_Experiment",
                   "gap_voltage_control": "_GapVoltageControl",
                   "pll": "_PLLControl",
                   "piezo_control": "_PiezoControl",
                   "regulator": "_Regulator",
                   "view": "_View",
                   "xy_scanner": "_XYScanner",
                   "spectroscopy": "_Spectroscopy"}

    if module_name in all_modules.keys():
        mate_name = all_modules[module_name]
    elif module_name in all_modules.values():
        mate_name = module_name
    else:
        raise ValueError(f"Requested module {module_name} must be one of {all_modules.keys()}")

    return mate_name


def is_parameter_allowable(value, experiment_element, parameter, test=0):
    """
    Checks if a parameter is allowable within the reported operating range of the equipment.

    Using this before sending an out of range prevents possible locking up/hard crashes of Matrix.

    Attributes
    ----------
    experiment_element : str
        The experiment element that the desired parameter falls under
    parameter : str
        The parameter to be tested
    value :
        The value of the parameter

    Returns
    -------
    response : bool

    Examples
    --------
    >>> is_parameter_allowable(100, "xy_scanner", "Points")

    """
    if value is None:
        return True
    else:
        min_max = read_min_max(experiment_element, parameter, test)
        if min_max:
            try:
                response = min_max[0] <= value <= min_max[1]
            except:
                raise MatrixInvalidDataTypeError
            if not response:
                warnings.warn(
                    f"{parameter} ({value}) should be within range {min_max[0]} <= {parameter} <= {min_max[1]}. Matrix may die",
                    MatrixParameterOutOfRangeWarning)
        else:
            response = None
    return response


def read_min_max(experiment_element, parameter, test=0):
    """
    Reads the minimum and maximum allowed values for settable parameters.

    Attributes
    ----------
    experiment_element : str
        The experiment element that the desired parameter falls under
    parameter : str
        The parameter to be tested

    Returns
    -------
    outs : tuple
        A tuple in the form [min, max]

    Examples
    --------
    >>> read_min_max("xy_scanner", "Points")
    """

    p = 'function'
    experiment_element = _friendly_name_to_mate(experiment_element)
    members = inspect.getmembers(sys.modules['nOmicron.mate.objects'], inspect.isclass)
    inspected = [item[1] for item in members if item[0] == experiment_element][0]

    def get_value(min_max):
        if experiment_element == "_Clock":
            out = mo._process(p, [inspected(1.0), min_max], parameter, test)
        else:
            out = mo._process(p, [inspected(), min_max], parameter, test)
        if type(out) is tuple:
            out = out[0]
        return out

    outs = [get_value("min"), get_value("max")]
    return outs


def restore_z_functionality():
    mo.xy_scanner.X_Trace_Trigger(False)
    mo.xy_scanner.X_Retrace_Trigger(False)
    mo.xy_scanner.Y_Retrace_Trigger(False)
    mo.experiment.stop()


def get_allowed_channels():
    experiment_name = mo.mate.scope
    install_path = mo.mate.installation_directory
    experiment_path = f"{install_path}\Templates\default\Experiments\{experiment_name}.expd"

    experiment_file = open(experiment_path, "r").read()
    soup = BeautifulSoup(experiment_file, "xml")

    raw_entries = soup.findAll(panelType="ChannelControl")
    allowed_channels = [line.attrs["experimentElementInstanceName"] for line in raw_entries]

    return allowed_channels



# Oliver Gordon, 2019
import matplotlib.pyplot as plt
from matplotlib import colors
from matplotlib_scalebar.scalebar import ScaleBar

_nanomap_dict = {'red': ((0.0, 0.0, 0.0),
                         (0.5, 1.0, 1.0),
                         (1.0, 1.0, 1.0)),
                 'green': ((0.0, 0.0, 0.0),
                           (1.0, 1.0, 1.0)),
                 'blue': ((0.0, 0.0, 0.0),
                          (0.5, 0.0, 0.0),
                          (1.0, 1.0, 1.0))}  # As defined by SPIW
nanomap = colors.LinearSegmentedColormap('Nanomap', _nanomap_dict)


def plot_xy(xydata, title=None, pixel_scale=None):
    """Takes the CURRENT values from Matrix for the scalebar."""

    def plot(data, dir):
        plt.figure()
        plt.imshow(data, cmap=nanomap)

        plt.axis('off')
        plt.colorbar()
        if pixel_scale is not None:
            plt.gca().add_artist(ScaleBar(pixel_scale, units="nm"))
        if title is not None:
            plt.title(f"{title} {dir}")

    if len(xydata) == 2:
        plot(xydata[0], dir="Up")
        plot(xydata[1], dir="Down")
    else:
        plot(xydata, dir="Up")

    plt.show()


def plot_linear_signal(x, y, channel=None, title=None):
    """Plots a linear signal or set of spectroscopy curves, such as I(t) or I(V)"""

    fig = plt.figure()
    if isinstance(y, list):
        plt.close(fig)
        fig, axes = plt.subplots(nrows=len(y), ncols=1, sharex=True, sharey=True)
        for repeat in range(len(y)):
            if isinstance(y[0], list):
                axes[repeat].plot(x, y[repeat][0], 'green')
                axes[repeat].plot(x, y[repeat][1], 'gold')
                fig.legend(['Forward', 'Back'])
            else:
                axes[repeat].plot(x, y[repeat], 'green')
    else:
        plt.plot(x, y, 'green')

    if title is not None:
        plt.title(title)

    if channel is not None:
        plt.xlabel(channel[2])
        plt.ylabel(channel[0])



# Oliver Gordon, 2019
from time import sleep
import numpy as np
from nOmicron.mate import objects as mo


def tip_pulse(voltage, time, num_pulses=1, pos=None, feedback_loop=False):
    """
    Performs a tip voltage pulse.

    Parameters
    ----------
    voltage : float
        The voltage (in volts)
    time : float
        The time (in seconds)
    num_pulses : int
        The number of times to pulse. Default is 1
    pos : None or tuple
        The position to do the pulse at, in normalised co-ords. If None, do it at the current tip position
    feedback_loop : bool
        If the feedback loop is on or not. Default False
    """

    mo.xy_scanner.Execute_Port_Colour("VPulse")
    if np.abs(voltage) >= 1:
        mo.gap_voltage_control.Tip_Cond_Pulse_Preamp_Range(1)
    else:
        mo.gap_voltage_control.Tip_Cond_Pulse_Preamp_Range(0)

    old_feedback_loop = mo.regulator.Feedback_Loop_Enabled()
    mo.regulator.Feedback_Loop_Enabled(feedback_loop)

    mo.gap_voltage_control.Tip_Cond_Pulse_Time(time)
    mo.gap_voltage_control.Tip_Cond_Pulse_Voltage(voltage)

    if pos is not None:
        # Do a pulse at position
        mo.xy_scanner.Store_Current_Position(True)
        mo.xy_scanner.Target_Position(pos)
        for i in range(num_pulses):
            mo.experiment.pause()
            mo.xy_scanner.Trigger_Execute_At_Target_Position(True)
            mo.xy_scanner.move()
            mo.xy_scanner.Trigger_Execute_At_Target_Position(False)
            mo.experiment.resume()
        mo.xy_scanner.Return_To_Stored_Position(True)
        mo.xy_scanner.Store_Current_Position(False)
        sleep(0.01)
    else:
        # Do a time period pulse - Tip_Cond_Pulse_Apply is super unreliable (?)
        for i in range(num_pulses):
            old_voltage = mo.gap_voltage_control.Voltage()
            mo.gap_voltage_control.Voltage(voltage)
            sleep(time)
            mo.gap_voltage_control.Voltage(old_voltage)
            # sleep(0.01)

    mo.regulator.Feedback_Loop_Enabled(old_feedback_loop)


def tip_crash(delta_z, pos=(-1, -1), delay=0, slew_rate=None):
    """
    Performs a controlled tip indendation.

    Parameters
    ----------
    delta_z : float
        The depth to ramp in meters.
    pos : tuple
        The position in normalised co-ordinate range(-1, 1) to perform the crash at.
        Default is (-1, -1), which corresponds to the bottom left corner of the scan window
    delay : float
        The time to continuously sample Z position before doing the crash. The average value is used as the start
        position to calculate delta_z from. Default is 0
    slew_rate : float or None
        The slew rate in metres/second. Not enabled in None (default)

    """
    assert delta_z > 0, "delta_z must be positive"
    # mo.experiment.pause()
    mo.xy_scanner.Execute_Port_Colour("ZRamp")
    mo.regulator.Enable_Z_Ramp_Slew_Rate(False)
    mo.regulator.Z_Ramp_Delay(delay)

    if slew_rate is not None:
        mo.regulator.Enable_Z_Ramp_Slew_Rate(True)
        mo.regulator.Z_Ramp_Slew_Rate(slew_rate)
    mo.regulator.Z_Ramp(-delta_z)

    mo.xy_scanner.Execute_Port_Colour("ZRamp")
    mo.xy_scanner.Store_Current_Position(True)
    mo.xy_scanner.Target_Position(pos)
    mo.xy_scanner.Trigger_Execute_At_Target_Position(True)

    mo.xy_scanner.move()

    mo.xy_scanner.Trigger_Execute_At_Target_Position(False)
    mo.xy_scanner.Return_To_Stored_Position(True)
    mo.xy_scanner.Store_Current_Position(False)
    mo.xy_scanner.Target_Position()
    mo.xy_scanner.Execute_Port_Colour("")
    # mo.experiment.resume()


def tip_scratch(delta_z, end_pos, start_pos=None):
    """
    Super destructive. Press the tip into the surface, move it, then come back out.

    Attributes
    ----------
    delta_z : float
        The depth to ramp in meters.
    start_pos : tuple or None
        The position in normalised co-ordinate range(-1, 1) to perform the initial crash at. Set to None (default)
        to start from the current position of the tip
    end_pos : tuple or None
        The position in normalised co-ordinate range(-1, 1) to finish the scratch at.
    """

    mo.xy_scanner.Store_Current_Position(True)
    mo.regulator.Feedback_Loop_Enabled(False)

    # Go to start position
    if start_pos is not None:
        mo.xy_scanner.Target_Position(start_pos)
        mo.xy_scanner.move()

    mo.regulator.Z_Offset(delta_z)  # Press in the tip

    mo.xy_scanner.Target_Position(end_pos)  # Move the tip
    mo.xy_scanner.move()

    mo.regulator.Z_Offset(0)  # Un-press the tip
    mo.regulator.Feedback_Loop_Enabled(True)

    # Return to start position
    mo.xy_scanner.Return_To_Stored_Position(True)
    mo.xy_scanner.Store_Current_Position(False)



from setuptools import setup, find_packages

setup(
    name='nOmicron',
    version='1.6.2',
    url='https://github.com/OGordon100/nOmicron',
    license='GPL v3.0',
    author='Oliver Gordon',
    author_email='oliver.gordon@nottingham.ac.uk',
    description="Python Control for Scienta Omicron Matrix",
    long_description="nOmicron package/API to automatically control Matrix through Python",
    packages=find_packages(),
    install_requires=['numpy', 'scipy', 'tqdm', 'matplotlib', 'natsort', 'pefile', 'psutil', 'selenium', 'chromedriver_autoinstaller', 'bs4', 'lxml'],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: Microsoft :: Windows"]
)



#   Code up to mate4dummies 0.4.2 (at time of branch) is Copyright © 2015 - 2018 Stephan Zevenhuizen
#   MATE, (20-08-2018).
#   Additional changes by Oliver Gordon, 2019
import ctypes
import os
import re
import sys
import time
import xml.etree.ElementTree as ET

import pefile
import psutil
from natsort import natsort

from nOmicron.utils import errors


class MATE(object):
    class ValueType:
        vt_INTEGER = 1
        vt_DOUBLE = 2
        vt_STRING = 3
        vt_BOOLEAN = 4
        vt_ENUM = 5
        vt_PAIR = 6
        vt_ARRAY_OF_DOUBLE = 7

    class String(ctypes.Structure):
        _fields_ = [('length', ctypes.c_int),
                    ('text', ctypes.c_char * 256)]

    def flat_values(self, string_length, array_length, length):
        pack = 4 * (1 + int(self.machine == 34404))

        class RealArray(ctypes.Structure):
            _pack_ = pack
            _fields_ = [('length', ctypes.c_int),
                        ('values', ctypes.c_double * array_length)]

        class FlatValue(ctypes.Structure):
            _pack_ = pack
            _fields_ = [('type', ctypes.c_int),
                        ('boolean', ctypes.c_int),
                        ('integer', ctypes.c_int),
                        ('enumeration', ctypes.c_int),
                        ('real', ctypes.c_double),
                        ('string', ctypes.POINTER(ctypes.POINTER(self.String))),
                        ('pairX', ctypes.c_double),
                        ('pairY', ctypes.c_double),
                        ('realArray',
                         ctypes.POINTER(ctypes.POINTER(RealArray)))]

        class FlatValues(ctypes.Structure):
            _fields_ = [('length', ctypes.c_int),
                        ('values', FlatValue * length)]

        flat_values = FlatValues(length)
        for i in range(length):
            o = flat_values.values[i]
            o.string = ctypes.pointer(ctypes.
                                      pointer(self.String(string_length)))
            o.realArray = ctypes.pointer(ctypes.
                                         pointer(RealArray(array_length)))
            flat_values.values[i] = o
        return flat_values

    def __init__(self, log, exit_handler2, testmode):
        self.log = log
        self.exit_handler2 = exit_handler2
        self.testmode = testmode
        self.rcs = dict(RMT_DEBUG=0x00000000,
                        RMT_SUCCESS=0x00000001,
                        RMT_UNEXPECTEDRESPONSE=0x00000002,
                        RMT_LIBNOTLOADABLE=0x00000004,
                        RMT_INTERNALFAILURE=0x00000006,
                        RMT_NOSERVER=(0x00000010 | 0x00000002),
                        RMT_NOEVENT=(0x00000020 | 0x00000003),
                        RMT_INCOMPATIBLEPROTOCOL=(0x00000010 | 0x00000004),
                        RMT_DISCONNECTED=(0x00000010 | 0x00000006),
                        RMT_UNSUPPORTEDREQ=(0x00000020 | 0x00000005),
                        RMT_UNKNOWNOBJECT=(0x00000020 | 0x00000002),
                        RMT_UNKNOWNPROPERTY=(0x00000020 | 0x00000004),
                        RMT_INVALIDTYPE=(0x00000020 | 0x00000006),
                        RMT_REJECTED=(0x00000020 | 0x00000008))
        self.experiments_directory = ''
        self.online = False
        self.is_ran_down = True
        self.rc = self.rcs['RMT_SUCCESS']
        pe = pefile.PE(sys.executable, fast_load=True)
        self.machine = pe.FILE_HEADER.Machine
        pe.close()

    def rc_key(self, rc):
        rc_key = list(self.rcs.keys())[list(self.rcs.values()).index(rc)]
        return rc_key

    def exit_handler(self, rc):
        if rc != self.rcs['RMT_SUCCESS'] and rc != self.rcs['RMT_NOEVENT']:
            self.exit_handler2()
            self.disconnect()
            self.log.AppendText('Active MATRIX experiment/project closed or '
                                'MATRIX software terminated.\n')

    def remote_access(self, p, rc):
        if rc == self.rcs['RMT_SUCCESS'] or rc == self.rcs['RMT_NOEVENT']:
            if len(p) > 2:
                p[2] = p[2].encode()
            if p[1] == 'getString':
                s = self.String(255)
                p_s = ctypes.pointer(s)
                rc = self.lib_mate.getStringPropertyByDesc(p[2], -1,
                                                           ctypes.byref(p_s))
                out = s.text[:].decode()
            elif p[1] == 'getBoolean':
                b = ctypes.c_char()
                rc = self.lib_mate.getBooleanProperty(p[2], -1,
                                                      ctypes.byref(b))
                out = bool(ord(b.value))
            elif p[1] == 'getInteger':
                if len(p) == 4:
                    i = p[3]
                else:
                    i = ctypes.c_int()
                rc = self.lib_mate.getIntegerProperty(p[2], -1,
                                                      ctypes.byref(i))
                out = i.value
            elif p[1] == 'getEnum':
                e = ctypes.c_int()
                rc = self.lib_mate.getEnumProperty(p[2], -1,
                                                   ctypes.byref(e))
                out = e.value
            elif p[1] == 'getDouble':
                d = ctypes.c_double()
                rc = self.lib_mate.getDoubleProperty(p[2], -1,
                                                     ctypes.byref(d))
                out = d.value
            elif p[1] == 'getDoubleArray':
                count = ctypes.c_int(len(p[0]))
                p_values = ctypes.pointer(list(map(ctypes.c_double, p[0])))
                rc = self.lib_mate.getDoubleArrayProperty(p[2], -1,
                                                          ctypes.byref(count),
                                                          ctypes.byref(p_values)
                                                          )
                out = p_values[0][:count]
            elif p[1] == 'getPair':
                d1 = ctypes.c_double()
                d2 = ctypes.c_double()
                rc = self.lib_mate.getPairProperty(p[2], -1,
                                                   ctypes.byref(d1),
                                                   ctypes.byref(d2))
                out = d1.value, d2.value
            elif p[1] == 'getEvent':
                prop = self.String(255)
                p_prop = ctypes.pointer(prop)
                value_count = ctypes.c_int()
                p_values = p[0][2]
                rc = self.lib_mate.getEntityEventByDesc(ctypes.byref(p_prop),
                                                        ctypes.
                                                        byref(value_count),
                                                        ctypes.
                                                        byref(p_values), 0)
                out = prop.text[:].decode(), value_count.value, p_values
            elif p[1] == 'trigger':
                rc = self.lib_mate.triggerProperty(p[2], -1)
                out = None
            elif p[1] == 'setString':
                rc = self.lib_mate.setStringProperty(p[2], -1,
                                                     p[3].encode())
                out = None
            elif p[1] == 'setBoolean':
                rc = self.lib_mate.setBooleanProperty(p[2], -1, p[3])
                out = None
            elif p[1] == 'setInteger':
                rc = self.lib_mate.setIntegerProperty(p[2], -1, p[3])
                out = None
            elif p[1] == 'setEnum':
                rc = self.lib_mate.setEnumProperty(p[2], -1, p[3])
                out = None
            elif p[1] == 'setDouble':
                d = ctypes.c_double(p[3])
                rc = self.lib_mate.setDoubleProperty(p[2], -1, d)
                out = None
            elif p[1] == 'setDoubleArray':
                count = len(p[3])
                p_value = ctypes.pointer(list(map(ctypes.c_double, p[3])))
                rc = self.lib_mate.setDoubleArrayProperty(p[2], -1,
                                                          count, p_value)
                out = None
            elif p[1] == 'setPair':
                d1 = ctypes.c_double(p[3])
                d2 = ctypes.c_double(p[4])
                rc = self.lib_mate.setPairProperty(p[2], -1, d1, d2)
                out = None
            elif p[1] == 'setObserved':
                rc = self.lib_mate.setObservedEntity(p[2], p[3])
                out = None
            elif p[1] == 'function':
                flat_value = p[0]
                p_args = p[3]
                rc = self.lib_mate.callFunctionByDesc(p[2],
                                                      ctypes.byref(flat_value),
                                                      ctypes.byref(p_args))
                out = flat_value
            else:
                rc = 0
                out = p[0]
            if (rc != self.rcs['RMT_SUCCESS'] and
                    rc != self.rcs['RMT_NOEVENT'] and
                    rc != self.rcs['RMT_UNKNOWNOBJECT']):
                out = p[0]
                self.log.AppendText('MATRIX error, response: ' +
                                    self.rc_key(rc) + '.\n')
        else:
            out = p[0]
        self.check_for_response_error(rc)
        return out, rc

    def deployment_parameter(self, scope, eei_name, dp_name):
        path = self.experiments_directory
        try:
            tree = ET.parse(os.path.join(path, scope + '.exps'))
            root = tree.getroot()
            tag = 'ExperimentElementInstance'
            p = re.compile(r'^{.*?}(.*)')
            elements = [i for i in root if p.match(i.tag).group(1) == tag]
            eei = [i for i in elements if i.get('name') == eei_name]
            if len(eei) == 1:
                dp = [i for i in eei[0] if i.get('name') == dp_name]
                if len(dp) == 1:
                    value = dp[0].get('value').split('::')[0]
                else:
                    value = ''
            else:
                value = ''
        except:
            value = ''
        return value

    def experiment(self):
        path = self.experiments_directory
        try:
            experiments = [s[:-5] for s in os.listdir(path)
                           if s.endswith('.expd')]
        except:
            experiments = []
        l = len(experiments)
        e = enumerate(experiments)
        rc = 1
        state = 'closed'
        flag = (l != 0)
        while flag:
            i = next(e)
            prop = i[1] + '::' + i[1] + '.State'
            func_params = ['', 'getString', prop]
            state, rc = self.remote_access(func_params, 1)
            flag = (((rc == self.rcs['RMT_SUCCESS']) or
                     (rc == self.rcs['RMT_UNKNOWNOBJECT'])) and i[0] < l - 1 and
                    ((state == 'closed') or (state == '')))
        if ((rc == self.rcs['RMT_SUCCESS']) and (state != 'closed')) or \
                self.testmode:
            self.scope = i[1]
            self.lib_mate.setScopeName(i[1].encode())
            name = 'testmode'
            rfn = time.strftime('%Y%m%d-%H%M%S', time.localtime()) + \
                  '_' + name
            rfp = os.path.join(os.environ['USERPROFILE'], 'SPM-data')
            if not os.path.exists(rfp):
                try:
                    os.mkdir(rfp)
                except:
                    pass
            self.exp_params = {'Name': name, 'Result_File_Name': rfn,
                               'Result_File_Path': rfp}
            for parameter in list(self.exp_params.keys()):
                if not self.testmode:
                    prop = self.scope + '.' + parameter
                    func_params = ['', 'getString', prop]
                    value, rc = self.remote_access(func_params, rc)
                    self.exp_params[parameter] = value
                self.log.AppendText('Experiment parameter ' + parameter + ': ' +
                                    self.exp_params[parameter] + '.\n')
            self.online = (rc == self.rcs['RMT_SUCCESS']) or self.testmode
        else:
            self.scope = ''
            self.log.AppendText('No open experiments found.\n')
            self.online = False
        self.check_for_response_error(rc)

    def connect(self):
        bin_sub_path = 'Bin\\Matrix.exe'
        library_sub_path = 'SDK\\RemoteAccess\\RemoteAccess_API.dll'
        name = os.path.basename(bin_sub_path)
        ok = False
        ps = psutil.process_iter()
        p = next(ps)
        go = True
        while go and not ok:
            try:
                p_path = p.exe()
                p_name = p.name()
            except:
                p_path = ''
                p_name = ''
            if p_name == name:
                installation_directory = p_path[:-(len(bin_sub_path) + 1)]
                library_path = os.path.join(installation_directory,
                                            library_sub_path)
                self.installation_directory = installation_directory
                self.library_path = library_path
                ok = os.path.exists(library_path)
            try:
                p = next(ps)
            except StopIteration:
                go = False
        ps.close()
        co = ''
        if ok:
            try:
                pe = pefile.PE(p_path)
                if pe.FILE_HEADER.Machine != self.machine:
                    self.log.AppendText('Bit architecture mismatch.\n')
                    pe.close()
                    pe = None
            except:
                pe = None
            if hasattr(pe, 'FileInfo') and pe.FileInfo:
                if isinstance(pe.FileInfo[0], list):
                    file_info = pe.FileInfo[0]
                else:
                    file_info = pe.FileInfo
                entries = [i for i in file_info if hasattr(i, 'StringTable')]
                if entries:
                    st_entries = [i for i in entries[0].StringTable]
                    if st_entries:
                        co = st_entries[0].entries[b'CompanyName'].decode()
                pe.close()
        if co:
            user_config_dir = os.environ['APPDATA']
            all_default_paths = natsort.natsorted(os.listdir(f"{user_config_dir}\\{co}\\MATRIX"))
            self.matrix_dir = all_default_paths[-1]
            exp_sub_path = f'MATRIX\\{all_default_paths[-1]}\\Experiments'
            self.experiments_directory = os.path.join(user_config_dir, co,
                                                      exp_sub_path)
            self.lib_mate = ctypes.cdll.LoadLibrary(library_path)
            self.library_path = library_path
            self.lib_mate.setHost(b'localhost')
            self.disconnect()
            if self.is_ran_down or self.testmode:
                rc = self.lib_mate.init(installation_directory.encode())
                self.log.AppendText('Connecting to the MATRIX, response: ' +
                                    self.rc_key(rc) + '.\n')
                if (rc == self.rcs['RMT_SUCCESS']) or self.testmode:
                    self.is_ran_down = False
                    self.experiment()
                    if not self.online:
                        self.disconnect()
                    else:
                        self.rc = self.rcs['RMT_SUCCESS']
                else:
                    self.rc = rc
        else:
            self.log.AppendText('Connecting to the MATRIX, response: '
                                '---.\n')

        self.check_for_response_error(self.rc)

    def disconnect(self):
        if not self.is_ran_down:
            rc = self.lib_mate.rundown()
            self.log.AppendText('Disconnecting from the MATRIX, response: ' +
                                self.rc_key(rc) + '.\n')
            self.is_ran_down = (rc == self.rcs['RMT_SUCCESS'])
            self.online = False
            self.rc = rc

            self.check_for_response_error(self.rc)

    def check_for_response_error(self, rc):
        rc_key = self.rc_key(rc)

        if rc_key == "RMT_UNEXPECTEDRESPONSE":
            raise errors.MatrixUnexpectedResponseError
        elif rc_key == "RMT_LIBNOTLOADABLE":
            raise errors.MateLibNotLoadableError
        elif rc_key == "RMT_INTERNALFAILURE":
            raise errors.MatrixCriticalHardwareFailureError
        elif rc_key == "RMT_NOSERVER":
            raise errors.MatrixNotInitialisedError
        elif rc_key == "RMT_INCOMPATIBLEPROTOCOL":
            raise errors.MatrixIncompatibleProtocolError
        elif rc_key == "RMT_UNSUPPORTEDREQ":
            raise errors.MatrixUnsupportedReqError
        elif rc_key in ["RMT_UNKNOWNOBJECT", "RMT_UNKNOWNPROPERTY"]:
            raise errors.MatrixUnsupportedOperationError
        elif rc_key == "RMT_INVALIDTYPE":
            raise errors.MatrixInvalidDataTypeError
        elif rc_key == "RMT_REJECTED":
            raise errors.MatrixRejectedError



# Oliver Gordon, 2019

import numpy as np

from nOmicron.mate import objects as mo
from nOmicron.microscope import IO
from tqdm import tqdm

def get_continuous_signal(channel_name, sample_time, sample_points):
    """Acquire a continuous signal.

    Parameters
    -----------
    channel_name : str
        The continuous channel to view, e.g. I_t, Z_t, Df_t, Aux1_t
    sample_time : float
        The time to acquire in seconds
    sample_points : int
        The number of points to acquire

    Returns
    -------
    x_data : Numpy array
    y_data : Numpy array

    Examples
    --------
    Acquire 60 points of I(t) data over 0.1 seconds
    >>> from nOmicron.microscope import IO
    >>> from nOmicron.utils.plotting import plot_linear_signal
    >>> IO.connect()
    >>> t, I = get_continuous_signal("I(t)", 1e-1, 60)
    >>> plot_linear_signal(v, I, "I(V)")
    >>> IO.disconnect()

    Acquire 100 points of Z(t) data over 5 seconds
    >>> from nOmicron.microscope import IO
    >>> from nOmicron.utils.plotting import plot_linear_signal
    >>> IO.connect()
    >>> t, I = get_continuous_signal("Z(t)", 5, 100)
    >>> plot_linear_signal(v, I, "I(V)")
    >>> IO.disconnect()
    """

    global view_count, x_data, y_data
    x_data = y_data = None
    view_count = 0

    def view_continuous_callback():
        global view_count, x_data, y_data
        view_count += 1
        pbar.update(1)

        data_size = mo.view.Data_Size()
        period = mo.clock.Period()
        x_data = np.linspace(0, (data_size - 1) * period, data_size)
        y_data = np.array(mo.sample_data(data_size))

    IO.enable_channel(channel_name)
    IO.set_clock(sample_time, sample_points)

    mo.view.Data(view_continuous_callback)
    mo.allocate_sample_memory(sample_points)

    pbar = tqdm(total=1)
    while view_count < 1 and mo.mate.rc == mo.mate.rcs['RMT_SUCCESS']:
        mo.wait_for_event()
    mo.clock.Enable(False)
    mo.view.Data()

    IO.disable_channel()

    return x_data, y_data


def get_point_spectra(channel_name, target_position, start_end, sample_time, sample_points,
                      repeats=1, forward_back=True, return_filename=False):
    """
    Go to a position and perform fixed point spectroscopy.

    Parameters
    ----------
    channel_name : str
        The channel to acquire from, e.g. I_V, Z_V, Aux2_V
    target_position : list
        [x, y] in the range -1,1. Can be converted from real nm units with utils.convert...
    start_end : tuple
        Start and end I/Z/Aux2
    sample_time : float
        The time to acquire in seconds
    sample_points : int
        The number of points to acquire
    repeats : int
        The number of repeat spectra to take for each point
    forward_back : bool
        Scan in both directions, or just one.
    return_filename : bool, optional
        If the full file name of the scan should be returned along with the data. Default is False

    Returns
    -------
    x_data : Numpy array
    y_data :
        If performing repeat spectra and:
            Scanning in both directions: list of list of Numpy arrays, where inner list is [0] forwards, [1] backwards
            Scanning in one direction: list of Numpy arrays
        If no repeat spectra and:
            Scanning in both directions: list of Numpy arrays, where list is [0] forwards, [1] backwards
            Scanning in one direction: single Numpy array

    Examples
    --------
    Acquire 60 points of I(V) data over 10 milliseconds, with tip placed in middle of scan window.
    >>> from nOmicron.microscope import IO
    >>> from nOmicron.utils.plotting import plot_linear_signal
    >>> IO.connect()
    >>> v, I = get_point_spectra("I(V)", start_end=[0, 1], target_position=[0, 0], ...
    >>>                   repeats=3, sample_points=50, sample_time=10e-3, forward_back=True)
    >>> plot_linear_signal(v, I, "I(V)")
    >>> IO.disconnect()
    """

    global view_count, view_name, x_data, y_data
    modes = {"V": 0, "Z": 1, "Varied Z": 2}  # Varied Z not fully supported yet!
    max_count = (repeats * (forward_back + 1))
    view_count = 0
    x_data = None
    y_data = []
    [y_data.append([None] * (bool(forward_back) + 1)) for i in range(repeats)]  # Can't use [] ** repeats

    def view_spectroscopy_callback():
        global view_count, view_name, x_data, y_data
        pbar.update(1)
        view_count += 1
        view_name = [mo.view.Run_Count(), mo.view.Cycle_Count()]
        cycle_count = mo.view.Cycle_Count() - 1
        packet_count = mo.view.Packet_Count() - 1
        data_size = mo.view.Data_Size()
        x_data = np.linspace(start_end[0], start_end[1], data_size)
        y_data[cycle_count][packet_count] = np.array(mo.sample_data(data_size))*1e-9
        if packet_count == 1:
            y_data[cycle_count][packet_count] = np.flip(y_data[cycle_count][packet_count])

    # Set all the parameters
    IO.enable_channel(channel_name)
    mo.spectroscopy.Spectroscopy_Mode(modes[channel_name[-2]])
    getattr(mo.spectroscopy, f"Device_{modes[channel_name[-2]] + 1}_Points")(sample_points)
    getattr(mo.spectroscopy, f"Raster_Time_{modes[channel_name[-2]] + 1}")(sample_time)
    getattr(mo.spectroscopy, f"Device_{modes[channel_name[-2]] + 1}_Start")(start_end[0])
    getattr(mo.spectroscopy, f"Device_{modes[channel_name[-2]] + 1}_End")(start_end[1])
    getattr(mo.spectroscopy, f"Device_{modes[channel_name[-2]] + 1}_Repetitions")(repeats)
    getattr(mo.spectroscopy, f"Enable_Device_{modes[channel_name[-2]] + 1}_Ramp_Reversal")(forward_back)

    # Set up spec
    mo.xy_scanner.Store_Current_Position(True)
    mo.xy_scanner.Target_Position(target_position)
    mo.xy_scanner.Trigger_Execute_At_Target_Position(True)

    # Do it
    mo.xy_scanner.move()
    mo.allocate_sample_memory(sample_points)
    mo.view.Data(view_spectroscopy_callback)

    pbar = tqdm(total=max_count)
    while view_count < max_count and mo.mate.rc == mo.mate.rcs['RMT_SUCCESS']:
        mo.wait_for_event()
    mo.view.Data()

    # Return to normal
    mo.xy_scanner.Trigger_Execute_At_Target_Position(False)
    mo.xy_scanner.Return_To_Stored_Position(True)
    mo.xy_scanner.Store_Current_Position(False)
    IO.disable_channel()

    if not forward_back:
        y_data = [item[0] for item in y_data]
    if repeats == 1:
        y_data = y_data[0]

    if return_filename:
        filename = f"{mo.experiment.Result_File_Path()}\\{mo.experiment.Result_File_Name()}--{view_count[0]}_{view_count[1]}.{channel_name}_mtrx"
        return x_data, y_data, filename
    else:
        return x_data, y_data


if __name__ == '__main__':
    from nOmicron.utils.plotting import plot_linear_signal
    IO.connect()
    t, I1 = get_continuous_signal("I(t)", sample_time=5, sample_points=50)

    # Do a fixed point spec
    v, I2 = get_point_spectra("I(V)", start_end=(0.5, -0.5), target_position=[0, 0.5],
                              repeats=4, sample_points=100, sample_time=20e-3, forward_back=True)
    plot_linear_signal(v, I2, "I(V)")
    IO.disconnect()



from nOmicron.mate import objects as mo

mo.pll.PLL_Enable(True)


def enable_pll():
    mo.pll.PLL_Enable(True)


def disable_pll():
    mo.pll.PLL_Enable(False)


def is_pll_locked():
    """Check if the PLL is locked"""
    return mo.pll.PLL_Locked()


def set_attenutation_level(attenuation):
    """
    Set the attenuation level

    Parameters
    ----------
    attenuation : int
        The level of attenuation from 0-4 (corresponds to 0, 0.1, 0.01, 0.001).
    """
    if mo.pll.Non_Contact_Mode() == 0:
        mo.pll.Excitation_Attenuation_Constant_Amplitude(attenuation)
    elif mo.pll.Non_Contact_Mode() == 1:
        mo.pll.Excitation_Attenuation_Constant_Excitation(attenuation)
    elif mo.pll.Non_Contact_Mode() == 2:
        mo.pll.Excitation_Attenuation_Self_Excitation(attenuation)


def set_loop_gain(I, P, method='new'):
    """Sets the loop gain"""

    if method == 'new':
        mo.pll.PLL_Loop_Gain_I(I)
        mo.pll.PLL_Loop_Gain_P(P)
    else:
        if mo.pll.Non_Contact_Mode() == 0:
            mo.pll.Amplitude_Loop_Gain_I_Constant_Amplitude(I)
            mo.pll.Amplitude_Loop_Gain_P_Constant_Amplitude(P)
        elif mo.pll.Non_Contact_Mode() == 1:
            raise BlockingIOError("Cannot set loop gain for constant excitation PLL mode")
        elif mo.pll.Non_Contact_Mode() == 2:
            mo.pll.Amplitude_Loop_Gain_I_Self_Excitation(I)
            mo.pll.Amplitude_Loop_Gain_P_Self_Excitation(P)
